//////////////////////////////////////////////////////////////////////
// 
// File:        LSoundPlayer.cpp
// Author:      Brian Postma
// Created:     30 December 1999
//
// Purpose:     The hardware/OS dependent soundplayer layer. Provides
//              all means and methods to play sounds.
//
//////////////////////////////////////////////////////////////////////

#include "LSoundPlayer.h"

LSoundPlayer::LSoundPlayer(char *sounddev, Module *modptr)
{
  this->modptr=modptr;
  soundhandle=-1;
  channels=NULL;
  nrofvoices=0;
  stereo=false;
  playbuffer=NULL;
  sounddevice=new char[strlen(sounddev)];
  if (sounddevice==NULL)
    return;
  strcpy(sounddevice,sounddev);
}


LSoundPlayer::~LSoundPlayer()
{
  CleanUp();
}


void LSoundPlayer::CleanUp()
{
  delete[] sounddevice;
  if (soundhandle!=-1)
    close(soundhandle);
  delete[] channels;
  delete[] playbuffer;
}


bool LSoundPlayer::Initialize(short nrofvoices, bool stereo)
{
  int		fragment = FRAGMENT;
  int		format   = FORMAT;
  int		rate     = SAMPLERATE;

  if (sounddevice==NULL || modptr==NULL)
    return false;
  this->nrofvoices=nrofvoices;
  if (stereo)
    this->stereo=1;
  else
    this->stereo=0;
  if ((soundhandle=open(sounddevice,O_WRONLY | O_NONBLOCK))==-1)
    return false;
  if (ioctl(soundhandle,SNDCTL_DSP_SETFRAGMENT,&fragment)==-1)
    return false;
  if (ioctl(soundhandle,SNDCTL_DSP_STEREO,&this->stereo)==-1)
    return false;
  if (ioctl(soundhandle,SNDCTL_DSP_SETFMT,&format)==-1)
    return false;
  if (ioctl(soundhandle,SNDCTL_DSP_SPEED,&rate)==-1)
    return  false;
  if ((channels=new LChannel[nrofvoices])==NULL)
    return false;
  if ((playbuffer=new BYTE[BUFFERSIZE*2])==NULL)
    return false;
  return true;
}


bool LSoundPlayer::Play(short voicenr, BYTE instrument,
                        int frequency, int volume)
{
  WORD size, repeat, replen,dummy;
  BYTE *sampptr;

  if (instrument<1 || instrument>15 || voicenr>=nrofvoices || voicenr<0)
    return false;
  if (volume>64) volume=64;
  if (modptr->IsSynthInstrument(instrument))
  {
    modptr->GetSynthInstrument(instrument,size,sampptr);
    replen=size;
    repeat=0;
  }
  else
  {
    modptr->GetSampledInstrument(instrument,size,repeat,replen,dummy,sampptr);
  }
  channels[voicenr].SetPlayBuffer(sampptr,size,repeat,replen);
  channels[voicenr].SetVolume(volume);
  channels[voicenr].SetFrequency(frequency);
  channels[voicenr].StartPlay();
  return true;
}


bool LSoundPlayer::Stop(short voicenr)
{
  if (voicenr<0 || voicenr>=nrofvoices)
    return false;
  channels[voicenr].StopPlay();
  return true;
}


bool LSoundPlayer::SetPan(short voicenr, int val)
{
  if (voicenr<0 || voicenr>=nrofvoices)
    return false;
  if (val<0 || stereo==0) 
    channels[voicenr].SetLeftRight(true);
  else
    channels[voicenr].SetLeftRight(false);
  return true;
}


bool LSoundPlayer::SetFrequency(short voicenr, int frequency)
{
  if (voicenr<0 || voicenr>=nrofvoices)
    return false;
  channels[voicenr].SetFrequency(frequency);
  return true;
}


bool LSoundPlayer::SetVolume(short voicenr, int volume)
{
  if (voicenr<0 || voicenr>=nrofvoices)
    return false;
  if (volume>64) volume=64;
  channels[voicenr].SetVolume(volume);
  return true;
}


bool LSoundPlayer::Update(short voicenr, BYTE *memptr, 
                          int offset, int numofbytes)
{
  if (voicenr<0 || voicenr>=nrofvoices)
    return false;
  channels[voicenr].Update(memptr,offset,numofbytes);
  return true;
}


bool LSoundPlayer::Mix()
{
  int left=0, right=0;
  BYTE lefttmp, righttmp;
  BYTE *ptr=playbuffer;
  int  towrite;
  int  written;

  for (int j=0; j<BUFFERSIZE; j++)
  {
    for (short i=0; i<nrofvoices; i++)
    {
      channels[i].GetNextBytes(lefttmp, righttmp);
      left+=(lefttmp-128);
      right+=(righttmp-128);
    }
    left=left/nrofvoices;
    right=right/nrofvoices;
    if (stereo)
    {
      *(ptr++)=left+128;
      *(ptr++)=right+128;
    }
    else
      *(ptr++)=left+128;
  }

  if (stereo)
    towrite=BUFFERSIZE*2;
  else
    towrite=BUFFERSIZE;
  ptr=playbuffer;

  while(towrite>0)
  {
    written=write(soundhandle,ptr,towrite);
    towrite-=written;
    ptr+=written;
    if (towrite>0)
    {
      usleep(1000);
      if ((written&1)!=0)
      {
        ptr++;
        towrite--;
      }
    }
  }    
  return true;
}
